#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: December 29, 2020
#
"""These files are for implementing Student-:math:`t` process regression.
It is implemented, based on the following article:

(i) Rasmussen, C. E., & Williams, C. K. (2006). Gaussian Process
Regression for Machine Learning. MIT Press.

(ii) Shah, A., Wilson, A. G., & Ghahramani, Z. (2014). Student-t Processes
as Alternatives to Gaussian Processes. In Proceedings of the 17th
International Conference on Artificial Intelligence and Statistics
(pp. 877-885)."""
