#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: March 22, 2021
#
"""These files are for implementing Gaussian process regression.
It is implemented, based on the following article:

(i) Rasmussen, C. E., & Williams, C. K. (2006). Gaussian Process
Regression for Machine Learning. MIT Press."""
